"""Console script for baseten."""

import configparser
import functools
import logging
import platform
import sys
import time

import click
import requests
from requests.exceptions import RequestException
from truss.cli import cli_group as truss_commands

from baseten import version as b10_version
from baseten.client_commands import (
    baseten_cli,
    dataset_cli,
    finetuning_cli,
    models_cli,
    pretrained_cli,
)
from baseten.common.settings import (
    get_server_url,
    read_config,
    set_config_value,
    set_server_url,
)
from baseten.common.util import setup_logger

logger = logging.getLogger(__name__)

DEFAULT_NETWORK_ERROR_MESSAGE = (
    "Sorry, something went wrong on our end. Please try again in 5 minutes."
)
# Message to display after getting a response from the server
SIGNUP_ERROR_MESSAGE = (
    "\nIf you have already signed up, please visit "
    "https://app.baseten.co/blueprint/tokens to "
    "initiate a login."
)


def ensure_login(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        config = read_config()
        try:
            config.get("api", "api_key")
        except configparser.NoOptionError:
            click.echo("You must first run the `baseten login` cli command.")
            sys.exit()
        result = func(*args, **kwargs)
        return result

    return wrapper


@click.group(name="baseten", invoke_without_command=True)
@click.pass_context
@click.option(
    "-v",
    "--version",
    is_flag=True,
    show_default=False,
    default=False,
    help="Show baseten package version.",
)
def cli_group(ctx, version):
    setup_logger("baseten", logging.INFO)
    if not ctx.invoked_subcommand:
        if version:
            click.echo(b10_version())
        else:
            click.echo(ctx.get_help())


@cli_group.command()
@click.option("--server_url", prompt="Baseten server URL", help="URL to hosted Baseten solution.")
def configure(server_url):
    """Configure client to use on-prem hosted Baseten environment.

    !!! note

        The client works out of the box without any need to run this command.
        This command is only required if you have Baseten Enterprise hosted on-prem and is not
        required for most users.
    """
    if set_server_url(server_url):
        click.echo("Saved server URL.")
    else:
        click.echo("That is not a valid URL.")


@cli_group.command()
@click.option(
    "--api_key", prompt="Baseten API key", hide_input=True, help="Valid API Key from Baseten."
)
def login(api_key: str):
    """Authenticate user with Baseten using API Key."""
    set_config_value("api", "api_key", api_key)
    click.echo("Saved API key.")


@cli_group.command()
@click.option(
    "--email", prompt="Email Address", hide_input=False, help="The email to register as user with."
)
def signup(email: str):
    """Signup as new Blueprint user.

    The sign-up process:

    1. Create an API generator token
    2. Create a baseten account with the provided email address
    3. Use API generator token to poll for newly created API Key
    """
    server_url = get_server_url()

    try:
        api_generator_response = requests.post(
            server_url + "/api/create_api_generator_token",
            json={"device_name": platform.node(), "device_os": platform.platform()},
        )

        if api_generator_response.status_code != 200:
            click.echo(DEFAULT_NETWORK_ERROR_MESSAGE)
            return

        token = api_generator_response.json()["token"]

        signup_response = requests.post(
            server_url + "/api/signup_new_user",
            json={"email": email, "is_package_user": True, "token": token},
        )
    except RequestException:
        click.echo(DEFAULT_NETWORK_ERROR_MESSAGE)
        return

    if signup_response.status_code != 200:
        click.echo(DEFAULT_NETWORK_ERROR_MESSAGE)
        return

    signup_response_json = signup_response.json()

    if signup_response_json["success"]:
        click.echo("Please check your email for a confirmation")
        _poll_for_available_api_key(token)
    else:
        click.echo(signup_response_json["error"])
        click.echo(SIGNUP_ERROR_MESSAGE)


def _poll_for_available_api_key(token):
    SLEEP_BETWEEN_POLLING_SECONDS = 2
    TIMEOUT_SECONDS = 300  # Poll for up to 5 minutes

    # Divide the total timeout by the amount sleeping in loop to
    # compute total # of iterations
    for _ in range(int(TIMEOUT_SECONDS / SLEEP_BETWEEN_POLLING_SECONDS)):
        try:
            response = requests.get(get_server_url() + f"/api/api_generator_token/{token}")
        except RequestException:
            time.sleep(SLEEP_BETWEEN_POLLING_SECONDS)
            continue

        api_key = response.json().get("api_key")
        if api_key is not None:
            set_config_value("api", "api_key", api_key)

            click.echo("API Key Successfully Saved.")
            break
        else:
            time.sleep(SLEEP_BETWEEN_POLLING_SECONDS)

    else:
        click.echo(DEFAULT_NETWORK_ERROR_MESSAGE)


truss_commands.add_command(baseten_cli.deploy_from_directory)
# Add Truss command group
cli_group.add_command(truss_commands)
# Add pretrained models command group
cli_group.add_command(pretrained_cli.pretrained)
# Add models command group
cli_group.add_command(models_cli.models)
# Add dataset command group
cli_group.add_command(dataset_cli.dataset_cli)
# Add finetuning command group
cli_group.add_command(finetuning_cli.finetuning_cli)

if __name__ == "__main__":
    cli_group()



import base64
import json
from typing import Any


def _bin_transform_str(s: str, transform_fn) -> str:
    return transform_fn(s.encode("utf-8")).decode("utf-8")


class Base64Codec:
    def encode(self, s: str) -> str:
        return _bin_transform_str(s, base64.b64encode)

    def decode(self, s: str) -> str:
        return _bin_transform_str(s, base64.b64decode)


Base64 = Base64Codec()


def encode_base64_json(obj: Any) -> str:
    return Base64.encode(json.dumps(obj))



import configparser
import os
import shutil
import tempfile
from pathlib import Path
from threading import local
from urllib.parse import urlparse

SKLEARN = "sklearn"
KERAS = "keras"
TENSORFLOW = "tensorflow"
PYTORCH = "pytorch"
CUSTOM = "custom"

DEBUG = False
CONFIG_FILE_PATH = f"{Path.home()}/.baseten_config"

thread_data = local()


def _write_file(config: configparser.ConfigParser, file_path: str):
    """
    We use a tempfile to write the contents to prevent race conditions around
    writing this file, in cases where `baseten.login` is used programmatically.
    This ensures that writes to the config file are atomic.
    """
    with tempfile.NamedTemporaryFile(delete=False, mode="w") as tmp_file:
        with open(tmp_file.name, "w") as tmp_file_handler:
            config.write(tmp_file_handler)
        shutil.move(tmp_file.name, file_path)


def is_config_file_valid() -> bool:
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        # read() doesn't raise an exception if the file doesn't exist
        return os.path.exists(CONFIG_FILE_PATH)
    except configparser.ParsingError:
        return False


def set_config_defaults():
    config = configparser.ConfigParser()
    if is_config_file_valid():
        config.read(CONFIG_FILE_PATH)
    if not config.has_section("api"):
        config.add_section("api")
    config.set("api", "url_base", "https://app.baseten.co")
    _write_file(config, CONFIG_FILE_PATH)

    return config


def read_config():
    if not is_config_file_valid():
        set_config_defaults()
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    return config


def set_config_value(section: str, key: str, value: str):
    config = read_config()
    if section not in config.sections():
        config.add_section(section)
    config.set(section, key, value)
    _write_file(config, CONFIG_FILE_PATH)


#    with open(CONFIG_FILE_PATH, "w") as configfile:
#        config.write(configfile)


def get_server_url() -> str:
    config = read_config()
    try:
        return config.get("api", "url_base")
    except (configparser.NoSectionError, configparser.NoOptionError):
        config = set_config_defaults()
        return config.get("api", "url_base")


def _is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def set_jwt_auth_token(token: str):
    thread_data.jwt_auth_token = token


def get_jwt_auth_token():
    if hasattr(thread_data, "jwt_auth_token"):
        return getattr(thread_data, "jwt_auth_token")
    return None


def set_server_url(server_url):
    if not _is_valid_url(server_url):
        return False
    set_config_value("api", "url_base", server_url.strip("/"))
    return True



from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api, settings
from baseten.models.foundational_model import FoundationalModel
from baseten.models.util import (
    get_or_create_pretrained_model,
    requests_error_handling,
    upload_file_to_s3,
)


class Whisper(FoundationalModel):
    """
    Whisper is a state-of-the-art transcription model from OpenAI.
    It is trained on 680,000 hours of multilingual and multitask supervised data.

    **Examples:**

    Invocation with a URL

    ```python
    from baseten.models import Whisper
    model = Whisper()
    transcript = model("https://baseten.s3.amazonaws.com/whisper/whisper_test.wav")
    ```

    Invocation with a local file

    ```python
    from baseten.models import Whisper
    model = Whisper()
    transcript = model("whisper_test.wav")
    ```
    """

    def __init__(self):
        self._model = self._set_user_model()

    def status(self) -> str:
        return self._model.status

    def id(self) -> str:
        return self._model.id

    def _set_user_model(self) -> BasetenDeployedModel:
        """Creates internal BasetenDeployedModel object that points to users
        deployed Whisper model. If the user does not have a deployed model,
        we will create one for them.
        """
        model_version = get_or_create_pretrained_model("Whisper")
        return BasetenDeployedModel(
            model_id=model_version["model_id"],
            truss_spec_version=model_version["truss_spec_version"],
        )

    def __call__(self, path: str, **kwargs) -> dict:
        """Generate text from an audio file. Supports local file paths or URLs.

        Args:
            path (str): Path to audio file. Can be a local file path or a URL.

        Returns:
            transcript: Dictionary containing 3 keys: `"language"`, `"segments"`, and `"text"`.

                - `"language"` is a str representing the language of the audio file.
                - `"segments"` is a list of dicts representing segments of audio.
                Each contains `"start"` and `"end"` keys corresponding to the start and end
                times of the segment in seconds. Each also contains a `"text"` key which is
                the text generated for that segment.
                - `"text"` is a string containing the complete text generated for the audio file.
        """
        is_url = path.startswith("http")
        if not is_url:
            path = upload_file_to_s3(path)

        request_body = {
            "url": path,
        }

        # TODO: Whisper model needs to have updated spec that returns status
        with requests_error_handling():
            server_response = self._model.predict(request_body)

        return server_response

    @staticmethod
    def url():
        """
        Use this static method to get a URL to the Stable Diffusion model page
        in your Blueprint project, which contains information about the model.

        **Example:**

        ```python
        from baseten.models import Whisper

        Whisper.url()
        ```
        """
        try:
            blueprint_project_id = api.get_blueprint_projects()[0]["id"]
            return (
                f"{settings.get_server_url()}/blueprint/projects/"
                f"{blueprint_project_id}?est=community-model--whisper"
            )
        except IndexError:
            raise ValueError("User does not have any blueprint projects.")



from abc import ABC, abstractmethod


class FoundationalModel(ABC):
    """
    Base class for models served by the baseten.models package.
    """

    @property
    @abstractmethod
    def status(self):
        pass

    @property
    @abstractmethod
    def id(self):
        pass



from pathlib import Path
from typing import Optional

import click

from baseten.common.files import DatasetTrainingType, upload_dataset


@click.group(name="dataset")
def dataset_cli():
    """Manage datasets stored on Blueprint"""
    pass


@dataset_cli.command()
@click.argument("data_dir", type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option(
    "--name",
    "-n",
    required=False,
    type=str,
    help="Name given to Dataset on Blueprint. If none, the uploaded directory's name will be used instead.",
)
@click.option(
    "--training-type",
    "-t",
    required=True,
    type=click.Choice([t.value for t in DatasetTrainingType], case_sensitive=False),
    help="Types of FinetuningRun that can use this Dataset",
)
def upload(data_dir: Path, name: Optional[str], training_type: str):
    """Upload a new dataset to Baseten from DATA_DIR

    DATA_DIR is the path to directory with relevant files.
    """
    dataset_id = upload_dataset(
        path=Path(data_dir), training_type=DatasetTrainingType[training_type.upper()], name=name
    )
    click.echo()
    click.echo(f"Dataset ID:\n{dataset_id}")



import io
import json
import logging
import os
import tarfile
import tempfile
import time
from typing import Dict, List, Optional, Union

import requests
import yaml
from truss.truss_handle import TrussHandle

from baseten.common import api, settings
from baseten.common.core import raises_api_error
from baseten.common.error_handler import error_handling
from baseten.common.lib_support import (
    build_h5_data_object,
    coerce_input_to_json,
    version_gte,
)
from baseten.common.util import exists_model

logger = logging.getLogger(__name__)

REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC = 3
REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES = 20


@raises_api_error
def install_requirements(req_filepath: str):
    with open(req_filepath, "r") as fp:
        requirements_txt = fp.read()
    logger.info("🚀 Sending requirements to Baseten 🚀")
    resp = api.install_requirements(requirements_txt)
    status = resp["status"]
    if status == "PROCESSING":
        logger.info("🐳 Requirements are being installed 🐳")

        requirement_id = resp["id"]
        tries = 0
        while tries < REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES:
            time.sleep(REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC)
            resp = api.requirement_status(requirement_id)
            status = resp["status"]
            if status != "PROCESSING":
                break
            tries += 1
        else:
            logger.info(
                "⌛ Requirements are still being installed. Check the status by running "
                f"baseten.requirements_status('{requirement_id}') ⌛"
            )
    if status == "SUCCEEDED":
        logger.info("🖖 Installed requirements successfully 🖖")
    elif status == "FAILED":
        error_message = resp["error_message"]
        logger.info(f'⚠️ Failed to install requirements. Error: "{error_message}" ⚠️')


@raises_api_error
def requirements_status(requirement_id: str):
    return api.requirement_status(requirement_id)


class BasetenDeployedModel:
    """
    !!! hint
        `BasetenDeployedModel` requires exactly one of `model_id`, `model_version_id`
        or `external_model_version_id` during initialization.
    """

    @raises_api_error
    def __init__(
        self,
        model_id: Optional[str] = None,
        model_version_id: Optional[str] = None,
        external_model_version_id: Optional[str] = None,
        model_name: Optional[str] = None,
        model_zoo_name: Optional[str] = None,
        truss_handle: Optional[TrussHandle] = None,
        truss_spec_version: Optional[str] = None,
    ):
        """
        Args:
            model_id: The ID of a model from your Baseten account
            model_version_id: The ID of a model version from your Baseten account
            external_model_version_id: If you would rather track model IDs with your own system,
                pass the custom ID here
            model_name: The name of the deployed model
            model_zoo_name: Field for internal use, please ignore
            truss_handle: The Truss associated with the model
            truss_spec_version: The Truss specification for the model
        """
        if not model_id and not model_version_id and not external_model_version_id:
            raise ValueError(
                "Either model_id, model_version_id, or external_model_version_id must be provided."
            )

        if (
            model_id
            and (model_version_id or external_model_version_id)
            or (external_model_version_id and model_version_id)
        ):
            raise ValueError(
                "Must provide only one of model_id, model_version_id, or external_model_version_id; not more."
            )

        self._model_id = model_id
        self._model_version_id = model_version_id
        self._external_model_version_id = external_model_version_id
        if self._external_model_version_id:
            self._model_version_id = api.model_version_external_id_get_version(
                self._external_model_version_id
            )

        self._model_name = model_name
        self._truss_spec_version = truss_spec_version
        self.model_zoo_name = model_zoo_name
        self.truss_handle = truss_handle

    @property
    def model_version_id(self) -> Optional[str]:
        return self._model_version_id

    @property
    @raises_api_error
    def id(self) -> str:
        """
        The model id. If only model_version is supplied, fetches the model_id from
        the API.
        """
        if not self._model_id:
            # We add this intermediate variable to satisfy the type checks (self._model_id is Optional[str])
            model_id = api.model_id_from_model_version(self._model_version_id)
            self._model_id = model_id
        else:
            model_id = self._model_id

        return model_id

    @property
    def web_url(self) -> str:
        return f"{settings.get_server_url()}/models/{self._model_id}"

    @property
    def blueprint_url(self) -> str:
        try:
            blueprint_project_id = api.get_blueprint_projects()[0]["id"]
            return (
                f"{settings.get_server_url()}/blueprint/projects/"
                f"{blueprint_project_id}?est=deployed-model--{self._model_id}"
            )
        except IndexError:
            raise ValueError("User does not have any blueprint projects.")

    @property
    @raises_api_error
    def status(self):
        """Fetches the live status of the deployed model."""
        if self._model_version_id:
            return api.get_model_version_status(self._model_version_id)
        else:
            return api.get_primary_model_version_status(self._model_id)

    @raises_api_error
    @error_handling(error_prefix="Failed to invoke model.")
    def predict(
        self, inputs, metadata=None, prediction_only: bool = True
    ) -> Optional[Union[List[List], Dict]]:
        """Invokes the model given the input dataframe.

        Args:
            inputs (list | pandas.DataFrame | numpy.ndarray):
                The data representing one or more inputs to call the model with.
            metadata (Union[pd.DataFrame, List[Dict]]): Metadata key/value pairs (e.g. name, url), one for each input.
            prediction_only (bool): set to `False` for the full response (including probablities).

        Returns:
            A list of inferences for each given input; e.g.: [[3], [9]] would indicate the prediction for the
            first input in inputs_df is [3], and the prediction for the second is [9].

        Raises:
            TypeError: If the provided inputs is not of a supported type.
            ApiError: If there was an error communicating with the server.
        """
        if self._truss_spec_version is None:
            self._truss_spec_version = self.get_truss_spec_version()

        if not exists_model(
            model_name=None,
            model_id=self._model_id,
            model_version_id=self._model_version_id,
            models_provider=api.models,
        ):
            if self._model_version_id:
                logger.info(f"Could not find model with version id {self._model_version_id}.")
            else:
                logger.info(f"Could not find model with id {self._model_id}.")
            return None

        if version_gte(self._truss_spec_version, "2.0"):
            if self._model_version_id:
                return api.predict_for_model_version(self._model_version_id, inputs, v2=True)
            return api.predict_for_model(self._model_id, inputs, v2=True)

        inputs_list, metadata = coerce_input_to_json(inputs, metadata)

        if self._model_version_id:
            return api.predict_for_model_version(
                self._model_version_id, inputs_list, metadata, prediction_only
            )
        return api.predict_for_model(self._model_id, inputs_list, metadata, prediction_only)

    @raises_api_error
    def update_model_features(self, model_config_file_path: str):
        """Update the model's feature names and output class labels (if any) based on the config
        found at `model_config_file_path`

        Args:
            model_config_file_path (str): The path to the model config file
        """
        config_yaml = yaml.safe_load(open(model_config_file_path, "r"))
        feature_names = list(config_yaml["model_features"]["features"])
        class_labels = config_yaml.get("model_class_labels", [])
        api.update_model_features(self._model_version_id, feature_names, class_labels)

    @raises_api_error
    def set_primary(self):
        """Promote this version of the model as the primary version.
        Raises:
            ApiError: If there was an error communicating with the server.
        """
        if not self._model_version_id:
            raise ValueError(
                "Only a BasetenDeployedModel backed by a model_version can be set as primary."
            )
        return api.set_primary(self._model_version_id)

    @raises_api_error
    def deactivate(self):
        """Deactivate this version of the model.
        Raises:
            ApiError: If there was an error communicating with the server.
        """
        if not self._model_version_id:
            raise ValueError(
                "Only a BasetenDeployedModel backed by a model_version can be set as primary."
            )
        return api.deactivate_model_version(self._model_version_id)

    @raises_api_error
    def activate(self):
        """Activate this version of the model.
        Raises:
            ApiError: If there was an error communicating with the server.
        """
        if not self._model_version_id:
            raise ValueError(
                "Only a BasetenDeployedModel backed by a model_version can be set as primary."
            )
        return api.activate_model_version(self._model_version_id)

    @raises_api_error
    def upload_sample_data(
        self,
        feature_data,
        target_data=None,
        metadata: Optional[List[Dict]] = None,
        data_name: str = "validation_data",
    ) -> Dict:
        """Upload a subset of the training/validation data to be used for
            - Summary statistics for the model
            - To detect model drift
            - To use as baseline data for model interpretability
            - To seed new data in the client.

        Training and validation data with targets must be uploaded with the targets separate.

        Args:
            feature_data (Union[np.ndarray, pd.DataFrame, List[List]]): The feature data to upload.
            target_data (Union[np.ndarray, pd.DataFrame, List[List]]): The target data to upload.
            metadata (List[Dict]): Metadata key/value pairs for the dataset.
            data_name (str): The name of the data set.

        Returns:
            Dict: The status of the upload.
        """
        if not self._model_version_id:
            raise ValueError(
                "Please use on a BasetenDeployedModel instantiated with a model_version_id."
            )

        signed_s3_upload_post = api.signed_s3_upload_post(data_name)
        logger.debug(f"Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}")

        with tempfile.TemporaryDirectory() as data_temp_directory:
            data_temp_file = build_h5_data_object(
                feature_data, target_data, metadata, data_temp_directory
            )
            files = {"file": (f"{data_name}.h5", open(data_temp_file, "rb"))}
            form_fields = signed_s3_upload_post["form_fields"]
            form_fields["AWSAccessKeyId"] = form_fields.pop("aws_access_key_id")
            s3_key = form_fields["key"]
            requests.post(signed_s3_upload_post["url"], data=form_fields, files=files)
        return api.register_data_for_model(s3_key, self.model_version_id, data_name)

    def __repr__(self):
        attr_info = []
        if self._model_id:
            attr_info.append(f"model_id={self._model_id}")
        if self._model_version_id:
            attr_info.append(f"model_version_id={self._model_version_id}")
        if self._model_name:
            attr_info.append(f"name={self._model_name}")
        info_str = "\n  ".join(attr_info)
        return f"BasetenDeployedModel<\n  {info_str}\n>"

    @raises_api_error
    def get_truss_spec_version(self) -> dict:
        if self._model_id:
            return api.model_truss_spec_version(self._model_id)
        elif self._model_version_id:
            return api.model_version_truss_spec_version(self._model_version_id)
        elif self._external_model_version_id:
            return api.model_version_external_id_get_version(self._external_model_version_id)

        raise ValueError(
            "Please use on a BasetenDeployedModel instantiated with a model_version_id or model_id."
        )

    @raises_api_error
    def pull(self, directory="."):
        response = requests.get(api.get_model_s3_download_url(self.model_version_id))
        with tarfile.open(name=None, fileobj=io.BytesIO(response.content)) as t:
            t.extractall(os.path.join(directory, self.model_version_id))
        return response

    @property
    def truss(self) -> Optional[TrussHandle]:
        return self.truss_handle


@raises_api_error
def models_summary():
    return api.models_summary()
